/* GSKrabWin32Interface.m - this file is part of GSKrab
 *
 * Copyright (C) 2006 Wolfgang Sourdeau
 *
 * Author: Wolfgang Sourdeau <Wolfgang@Contre.COM>
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#define _WIN32_WINNT 0x0500

#import <windows.h>
#import <winuser.h>

#import <Foundation/NSAutoreleasePool.h>
#import <Foundation/NSDate.h>
#import <Foundation/NSString.h>
#import <Foundation/NSThread.h>

#import "GSKrabDesktopInterface.h"
#import "GSKrabManager.h"
#import "GSKrabWin32Interface.h"

@interface GSKrabWin32Interface (GSKrabPrivate)

- (BOOL) _handleKey: (int) win32KeyCode
          withNCode: (int) nCode
             wParam: (WPARAM) wParam
          andLParam: (LPARAM) lParam;

@end

static LRESULT CALLBACK
InterceptLLKeypress (int nCode, WPARAM wParam, LPARAM lParam)
{
  PKBDLLHOOKSTRUCT k;
  GSKrabWin32Interface *self;

  k = (PKBDLLHOOKSTRUCT) lParam;

  self = [GSKrabWin32Interface sharedInterface];

  return 

  return ((nCode == HC_ACTION
           && (wParam == WM_KEYDOWN || wParam == WM_SYSKEYDOWN)
           && [self _handleKey: win32KeyCode])
          || CallNextHookEx (keyboardHook, nCode, wParam, lParam));
}

@implementation GSKrabWin32Interface

+ (id) sharedInterface
{
  static GSKrabWin32Interface *sharedInterface = nil;

  if (!sharedInterface)
    sharedInterface = [self new];

  return sharedInterface;
}

- (BOOL) _handleKey: (int) win32KeyCode
{
  BOOL result;
  NSDictionary *parameters;
  id objects[2];
  id keys[] = { @"KrabKeyCode", @"returnPtr" };

  if (win32KeyCode < MAX_KEYS
      && handledKeys[win32KeyCode] != GSKrabNone)
    {
      objects[0] = [NSNumber numberWithInt: krabKeyCode];
      objects[1] = [NSValue valueWithPointer: &result];

      parameters = [NSDictionary dictionaryWithObjects: (id*)objects
                                 forKeys: (id*)keys
                                 count: 2];
      [self performSelectorOnMainThread: @selector (_forwardKeyPress:)
            withObject: parameters
            waitUntilDone: YES];
    }
  else
    result = NO;

  return result;
}

- (void) _forwardKeyPress: (NSDictionary *) parameters
{
  GSKrabKeyCode krabKeyCode;
  BOOL *result;

  krabKeycode = [[parameters objectForKey: @"KrabKeyCode"] intValue];
  result = [[parameters objectForKey: @"returnPtr"] pointerValue];

  *result = [manager handleKeyPress: krabKeyCode];
}

- (id) init
{
  Win32KeyEquivalent keyEquivalents[]
    = {{GSKrabPlay, VK_MEDIA_PLAY_PAUSE},
       {GSKrabStop, VK_MEDIA_STOP},
       {GSKrabTrackNext, VK_MEDIA_NEXT_TRACK},
       {GSKrabTrackPrevious, VK_MEDIA_PREV_TRACK},
       {GSKrabVolumeMute, VK_VOLUME_MUTE},
       {GSKrabVolumeIncrease, VK_VOLUME_UP},
       {GSKrabVolumeDecrease, VK_VOLUME_DOWN},
       {GSKrabEject, VK_LAUNCH_MEDIA_SELECT},
       {GSKrabWWW, VK_BROWSER_HOME},
       {GSKrabMail, VK_LAUNCH_MAIL},
       {GSKrabNone, -1}};
  Win32KeyEquivalent *keyEquivalent;

  if ((self = [super init]))
    {
      keyThreadId = 0;
      bindingTable = NSCreateMapTable (NSIntMapKeyCallBacks,
                                       NSIntMapValueCallBacks,
                                       GSKrabMaxNumberOfKeys);
      keyEquivalent = keyEquivalents;
      while (keyEquivalent->krabKeyCode != GSKrabNone)
        {
          NSMapInsert (bindingTable,
                       (void *) keyEquivalent->krabKeyCode,
                       (void *) keyEquivalent->win32KeyCode);
          keyEquivalent++;
        }
      memset (handledKeys, GSKrabNone, sizeof (GSKrabKeyCode) * MAX_KEYS);
    }

  return self;
}

- (void) setManager: (GSKrabManager *) krabManager
{
  manager = krabManager;
}

- (void) _Win32Loop
{
  MSG msg;
  NSAutoreleasePool *pool;

  pool = [NSAutoreleasePool new];

  keyThreadId = GetCurrentThreadId ();
  keyboardHook = SetWindowsHookEx (WH_KEYBOARD_LL,
                                   (HOOKPROC) InterceptLLKeypress,
                                   GetModuleHandle(NULL), 0L);

  while (GetMessage (&msg, NULL, 0, 0)
         && msg.message != WM_QUIT);
    {
      TranslateMessage (&msg);
      DispatchMessage (&msg);
    }

  UnhookWindowsHookEx (keyboardHook);
  keyThreadId = 0;

  [pool release];

  [NSThread exit];
}

- (void) start
{
  [NSThread detachNewThreadSelector: @selector (_Win32Loop)
            toTarget: self
            withObject: nil];
}

- (void) stop
{
  if (keyThreadId)
    PostThreadMessage (keyThreadId, WM_QUIT, 0, 0);
}

- (void) registerKeyCode: (GSKrabKeyCode) krabKeyCode
{
  int win32KeyCode;

  win32KeyCode = (int) NSMapGet (bindingTable, (void *) krabKeyCode);
  if (win32KeyCode)
    handledKeys[win32KeyCode] = krabKeyCode;
}

- (void) unregisterKeyCode: (GSKrabKeyCode) krabKeyCode
{
  int win32KeyCode;

  win32KeyCode = (int) NSMapGet (bindingTable, (void *) krabKeyCode);
  if (win32KeyCode)
    handledKeys[win32KeyCode] = GSKrabNone;
}

- (void) dealloc
{
  NSFreeMapTable (bindingTable);
  [super dealloc];
}

@end
